from __future__ import annotations
from copy import deepcopy
from enum import Enum, auto
import inspect

from typing import (
    TYPE_CHECKING,
    Callable,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    Literal,
)
from dataclasses import dataclass

import jax
import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
from numpyro.infer import MCMC, NUTS
from jax import Array
from jax.random import PRNGKey

from catalax.dataset.dataset import Dataset
from catalax.mcmc.results import HMCResults
from catalax.surrogate import Surrogate
from catalax.mcmc.protocols import PreModel, PostModel, Shapes

if TYPE_CHECKING:
    from catalax.model.model import Model


@dataclass
class MCMCConfig:
    """Configuration parameters for MCMC simulation.

    Args:
        num_warmup: Number of warmup steps before sampling
        num_samples: Number of posterior samples to collect
        likelihood: Likelihood distribution class (default: SoftLaplace)
        dense_mass: Whether to use dense mass matrix (default: True)
        thinning: Factor by which to thin samples (default: 1)
        max_tree_depth: Maximum depth of NUTS binary tree (default: 10)
        dt0: Time step resolution for simulation (default: 0.1)
        chain_method: Chain execution method - "sequential", "parallel", or "vectorized"
        num_chains: Number of Markov chains to run (default: 1)
        seed: Random seed for reproducibility (default: 420)
        verbose: Verbosity level; 0 for silent, 1 for progress (default: 1)
        max_steps: Maximum number of integration steps (default: 64^4)
    """

    num_warmup: int
    num_samples: int
    likelihood: Type[dist.Distribution] = dist.SoftLaplace
    dense_mass: bool = True
    thinning: int = 1
    max_tree_depth: int = 10
    dt0: float = 0.1
    chain_method: Literal["sequential", "parallel", "vectorized"] = "sequential"
    num_chains: int = 1
    seed: int = 420
    verbose: int = 1
    max_steps: int = 64**4
    solver: str = "scipy"


class Modes(Enum):
    """Modes of the model."""

    SURROGATE = auto()
    MECHANISTIC = auto()


@dataclass
class ModelSetup:
    """Container for model setup results."""

    priors: List[Tuple[str, dist.Distribution]]


@dataclass
class DataPreparation:
    """Container for prepared MCMC data."""

    data: Array
    times: Array
    y0s: Array
    constants: Array
    sim_func: Callable
    shapes: Shapes


class HMC:
    """Hamiltonian Monte Carlo sampler wrapper.

    This class provides a high-level interface for running HMC/NUTS sampling
    with integrated configuration and result handling.
    """

    def __init__(
        self,
        num_warmup: int,
        num_samples: int,
        likelihood: Type[dist.Distribution] = dist.SoftLaplace,
        dense_mass: bool = True,
        thinning: int = 1,
        max_tree_depth: int = 10,
        dt0: float = 0.1,
        chain_method: Literal["sequential", "parallel", "vectorized"] = "sequential",
        num_chains: int = 1,
        seed: int = 420,
        verbose: int = 1,
        max_steps: int = 64**4,
    ):
        """Initialize HMC sampler.

        Args:
            num_warmup: Number of warmup steps before sampling
            num_samples: Number of posterior samples to collect
            likelihood: Likelihood distribution class
            dense_mass: Whether to use dense mass matrix
            thinning: Factor by which to thin samples
            max_tree_depth: Maximum depth of NUTS binary tree
            dt0: Time step resolution for simulation
            chain_method: Chain execution method - "sequential", "parallel", or "vectorized"
            num_chains: Number of Markov chains to run
            seed: Random seed for reproducibility
            verbose: Verbosity level; 0 for silent, 1 for progress
            max_steps: Maximum number of integration steps
        """
        self.config = MCMCConfig(
            num_warmup=num_warmup,
            num_samples=num_samples,
            likelihood=likelihood,
            dense_mass=dense_mass,
            thinning=thinning,
            max_tree_depth=max_tree_depth,
            dt0=dt0,
            chain_method=chain_method,
            num_chains=num_chains,
            seed=seed,
            verbose=verbose,
            max_steps=max_steps,
        )
        self.likelihood = likelihood

    def run(
        self,
        model: "Model",
        dataset: Dataset,
        yerrs: Union[float, Array],
        surrogate: Optional[Surrogate] = None,
        pre_model: Optional[PreModel] = None,
        post_model: Optional[PostModel] = None,
    ) -> HMCResults:
        """Run MCMC simulation to infer posterior distribution of parameters.

        Uses NumPyro to perform Markov Chain Monte Carlo simulation with the No-U-Turn
        Sampler (NUTS) algorithm. Parameter priors are automatically extracted from the model.
        The simulation compares model output against observed data to infer the posterior
        distribution of parameters.

        Args:
            model: Model to fit
            dataset: Dataset containing observations to fit against
            yerrs: Standard deviation of observed data
            surrogate: Optional surrogate model for rate prediction
            pre_model: Optional function to transform inputs and parameters before simulation
            post_model: Optional function to transform outputs after simulation

        Returns:
            Tuple of (MCMC object with results, Bayesian model function)

        Raises:
            AssertionError: If any parameters lack prior distributions
            TypeError: If pre_model or post_model functions are not callable
            ValueError: If pre_model or post_model signatures don't match Protocol requirements
        """
        # Validate inputs
        _validate_parameter_priors(model)
        _validate_model_functions(pre_model, post_model)

        # Setup model and prepare data
        data_prep = _prepare_mcmc_data(dataset, model, surrogate)

        if surrogate is not None:
            mode = Modes.SURROGATE
        else:
            mode = Modes.MECHANISTIC

        # Create Bayesian model
        bayes_model = BayesianModel(
            model=model,
            yerrs=yerrs,
            likelihood=self.config.likelihood,
            sim_func=data_prep.sim_func,
            pre_model=pre_model,
            post_model=post_model,
            shapes=data_prep.shapes,
            mode=mode,
        )

        # Initialize and run MCMC
        nuts = NUTS(
            bayes_model,
            dense_mass=self.config.dense_mass,
            max_tree_depth=self.config.max_tree_depth,
        )
        mcmc = MCMC(
            nuts,
            num_warmup=self.config.num_warmup,
            num_samples=self.config.num_samples,
            progress_bar=bool(self.config.verbose),
            chain_method=self.config.chain_method,
            num_chains=self.config.num_chains,
            jit_model_args=True,
            thinning=self.config.thinning,
        )

        _run_mcmc_simulation(mcmc, data_prep, self.config)

        return HMCResults(mcmc, bayes_model, model)

    @classmethod
    def from_config(cls, config: MCMCConfig) -> HMC:
        """Create HMC instance from configuration.

        Args:
            config: MCMC configuration parameters
        """
        return cls(
            num_warmup=config.num_warmup,
            num_samples=config.num_samples,
            likelihood=config.likelihood,
            dense_mass=config.dense_mass,
            thinning=config.thinning,
            max_tree_depth=config.max_tree_depth,
            chain_method=config.chain_method,
            num_chains=config.num_chains,
            seed=config.seed,
            verbose=config.verbose,
            max_steps=config.max_steps,
            dt0=config.dt0,
        )


class BayesianModel:
    """Bayesian model for MCMC parameter inference.

    This class creates a callable Bayesian model that can be used with NumPyro's
    MCMC samplers. It encapsulates the model parameters, priors, and likelihood
    configuration.
    """

    def __init__(
        self,
        model: "Model",
        yerrs: Union[float, Array],
        likelihood: Type[dist.Distribution],
        sim_func: Callable,
        shapes: Shapes,
        mode: Modes,
        pre_model: Optional[PreModel] = None,
        post_model: Optional[PostModel] = None,
    ):
        """Initialize the Bayesian model.

        Args:
            model: The model being fit
            yerrs: Standard deviation of observed data
            likelihood: Likelihood distribution class
            sim_func: Function to simulate model with given parameters
            shapes: Shapes of the data, initial conditions, constants, and times
            pre_model: Optional function to transform inputs and parameters before simulation
            post_model: Optional function to transform outputs after simulation
        """
        self.model = deepcopy(model)
        self.model._setup_system()
        self.yerrs = yerrs
        self.likelihood = likelihood
        self.sim_func = sim_func
        self.pre_model = pre_model
        self.post_model = post_model
        self.shapes = shapes
        self.mode = mode

        # Pre-compute priors and observables
        self.priors = [
            (
                model.parameters[param].name,
                model.parameters[param].prior._distribution_fun(),
            )
            for param in model.get_parameter_order()
        ]

        self.observables = jnp.array(
            [
                i
                for i, species in enumerate(model.get_species_order())
                if model.odes[species].observable
            ]
        )

    def __call__(
        self,
        y0s: Array,
        constants: Array,
        times: Array,
        data: Optional[Array] = None,
    ):
        """Bayesian model for parameter posterior sampling.

        Samples parameters from priors, simulates the model, and compares
        against observed data to build the posterior distribution.

        Args:
            y0s: Initial conditions
            constants: System constants
            times: Time points for simulation
            data: Observed data to fit against

        Returns:
            Sampled posterior distribution
        """

        # Sample parameters from priors
        theta = jnp.array(
            [numpyro.sample(name, distribution) for name, distribution in self.priors]
        )

        # Apply a pre-model which can be used to modify
        # the initial conditions, constants, times, data, and parameters
        if self.pre_model is not None:
            y0s, constants, times, data, theta, pre_output = self.pre_model(
                model=self.model,
                y0s=y0s,
                constants=constants,
                times=times,
                data=data,
                theta=theta,
                shapes=self.shapes,
                mode=self.mode,
            )

        # Simulate model with sampled parameters
        states = self.sim_func(y0s, theta, constants, times)

        # Apply a post-model which can be used to modify
        # the states, data, and times
        if self.post_model is not None:
            states, data, times = self.post_model(
                model=self.model,
                states=states,
                data=data,
                times=times,
                pre_output=pre_output,
            )

        # Sample noise parameter
        sigma = numpyro.sample("sigma", dist.Normal(0, self.yerrs))  # type: ignore

        # Compare simulation to observed data
        numpyro.sample(
            "y", self.likelihood(states[..., self.observables], sigma), obs=data
        )


def run_mcmc(
    model: "Model",
    dataset: Dataset,
    yerrs: Union[float, Array],
    config: MCMCConfig,
    surrogate: Optional[Surrogate] = None,
    pre_model: Optional[PreModel] = None,
    post_model: Optional[PostModel] = None,
) -> HMCResults:
    """Run MCMC simulation to infer posterior distribution of parameters.

    Uses NumPyro to perform Markov Chain Monte Carlo simulation with the No-U-Turn
    Sampler (NUTS) algorithm. Parameter priors are automatically extracted from the model.
    The simulation compares model output against observed data to infer the posterior
    distribution of parameters.

    Args:
        model: Model to fit
        dataset: Dataset containing observations to fit against
        yerrs: Standard deviation of observed data
        config: MCMC configuration parameters
        surrogate: Optional surrogate model for rate prediction
        pre_model: Optional function to transform inputs and parameters before simulation
        post_model: Optional function to transform outputs after simulation

    Returns:
        Tuple of (MCMC object with results, Bayesian model function)

    Raises:
        AssertionError: If any parameters lack prior distributions
        TypeError: If pre_model or post_model functions are not callable
        ValueError: If pre_model or post_model signatures don't match Protocol requirements
    """
    # Validate inputs
    hmc = HMC.from_config(config)
    return hmc.run(model, dataset, yerrs, surrogate, pre_model, post_model)


def _prepare_mcmc_data(
    dataset: Dataset,
    model: "Model",
    surrogate: Optional[Surrogate],
) -> DataPreparation:
    """Prepare all data components needed for MCMC simulation.

    Args:
        dataset: Dataset containing observations
        model: Model to be fit
        config: MCMC configuration
        surrogate: Optional surrogate model

    Returns:
        DataPreparation containing all prepared data
    """
    # Extract basic dataset components
    data, times, y0s, constants = _extract_dataset_components(dataset, model)
    shapes = Shapes(
        y0s=y0s.shape,
        constants=constants.shape,
        times=times.shape,
        data=data.shape,
    )

    # Setup simulation function
    model = deepcopy(model)
    model._setup_system()
    sim_func, data, y0s, times, constants = _configure_simulation_function(
        model=model,
        surrogate=surrogate,
        dataset=dataset,
        data=data,
        y0s=y0s,
        times=times,
        constants=constants,
    )

    return DataPreparation(
        data=data,
        times=times,
        y0s=y0s,
        constants=constants,
        sim_func=sim_func,
        shapes=shapes,
    )


def _run_mcmc_simulation(
    mcmc: MCMC, data_prep: DataPreparation, config: MCMCConfig
) -> None:
    """Execute the MCMC simulation.

    Args:
        mcmc: Initialized MCMC sampler
        data_prep: Prepared data for simulation
        config: MCMC configuration
    """
    if config.verbose > 0:
        print("\n🚀 Running MCMC\n")

    mcmc.run(
        PRNGKey(config.seed),
        data=data_prep.data,
        y0s=data_prep.y0s,
        times=data_prep.times,
        constants=data_prep.constants,
    )

    if config.verbose > 1:
        mcmc.print_summary()


def _validate_parameter_priors(model: "Model") -> None:
    """Ensure all model parameters have specified prior distributions.

    Args:
        model: Model whose parameters should be validated

    Raises:
        AssertionError: If any parameters lack prior distributions
    """
    missing_priors = [
        param.name for param in model.parameters.values() if param.prior is None
    ]
    assert not missing_priors, (
        f"Parameters {', '.join(missing_priors)} do not have priors. Please specify priors for all parameters."
    )


def _validate_model_functions(
    pre_model: Optional[PreModel], post_model: Optional[PostModel]
) -> None:
    """Validate that pre_model and post_model functions conform to expected signatures.

    Args:
        pre_model: Optional pre-model transformation function
        post_model: Optional post-model transformation function

    Raises:
        TypeError: If functions don't conform to expected Protocol signatures
        ValueError: If functions have incorrect parameter signatures
    """
    if pre_model is not None:
        _validate_pre_model_signature(pre_model)

    if post_model is not None:
        _validate_post_model_signature(post_model)


def _validate_pre_model_signature(pre_model: PreModel) -> None:
    """Validate PreModel function signature.

    Args:
        pre_model: Pre-model function to validate

    Raises:
        TypeError: If function is not callable
        ValueError: If function signature doesn't match Protocol
    """
    if not callable(pre_model):
        raise TypeError("pre_model must be callable")

    try:
        sig = inspect.signature(pre_model)
    except (ValueError, TypeError) as e:
        raise ValueError(f"Could not inspect pre_model signature: {e}")

    # Expected parameters for PreModel
    expected_params = {"model", "y0s", "constants", "times", "data", "theta"}

    # Check that function accepts keyword arguments
    actual_params = set(sig.parameters.keys())

    # Check for **kwargs (accepts any keyword args) or exact match
    has_var_keyword = any(
        p.kind == inspect.Parameter.VAR_KEYWORD for p in sig.parameters.values()
    )

    if not has_var_keyword and not expected_params.issubset(actual_params):
        missing_params = expected_params - actual_params
        raise ValueError(
            f"pre_model missing required parameters: {missing_params}. "
            f"Expected signature: def pre_model(*, model, y0s, constants, times, data, theta)"
        )

    # Check that parameters are keyword-only (if not using **kwargs)
    if not has_var_keyword:
        for param_name, param in sig.parameters.items():
            if param_name in expected_params and param.kind not in (
                inspect.Parameter.KEYWORD_ONLY,
                inspect.Parameter.VAR_KEYWORD,
            ):
                raise ValueError(
                    f"pre_model parameter '{param_name}' must be keyword-only. "
                    f"Use: def pre_model(*, {param_name}, ...)"
                )


def _validate_post_model_signature(post_model: PostModel) -> None:
    """Validate PostModel function signature.

    Args:
        post_model: Post-model function to validate

    Raises:
        TypeError: If function is not callable
        ValueError: If function signature doesn't match Protocol
    """
    if not callable(post_model):
        raise TypeError("post_model must be callable")

    try:
        sig = inspect.signature(post_model)
    except (ValueError, TypeError) as e:
        raise ValueError(f"Could not inspect post_model signature: {e}")

    # Expected parameters for PostModel
    expected_params = {"model", "states", "data", "times"}

    # Check that function accepts keyword arguments
    actual_params = set(sig.parameters.keys())

    # Check for **kwargs (accepts any keyword args) or exact match
    has_var_keyword = any(
        p.kind == inspect.Parameter.VAR_KEYWORD for p in sig.parameters.values()
    )

    if not has_var_keyword and not expected_params.issubset(actual_params):
        missing_params = expected_params - actual_params
        raise ValueError(
            f"post_model missing required parameters: {missing_params}. "
            f"Expected signature: def post_model(*, model, states, data, times)"
        )

    # Check that parameters are keyword-only (if not using **kwargs)
    if not has_var_keyword:
        for param_name, param in sig.parameters.items():
            if param_name in expected_params and param.kind not in (
                inspect.Parameter.KEYWORD_ONLY,
                inspect.Parameter.VAR_KEYWORD,
            ):
                raise ValueError(
                    f"post_model parameter '{param_name}' must be keyword-only. "
                    f"Use: def post_model(*, {param_name}, ...)"
                )


def _extract_dataset_components(
    dataset: Dataset, model: "Model"
) -> Tuple[Array, Array, Array, Array]:
    """Extract and prepare dataset components for MCMC simulation.

    Args:
        dataset: Dataset containing observations
        model: Model to be fit to the data

    Returns:
        Tuple of (data, times, initial conditions, constants)
    """
    # Extract data, times, and initial conditions
    data, times, y0s = dataset.to_jax_arrays(
        model.get_species_order(),
        inits_to_array=True,
    )

    # Get constants from dataset
    constants = dataset.to_y0_matrix(species_order=model.get_constants_order())

    # Create initial conditions for all species, including non-observable ones
    all_species = model.get_species_order()
    full_y0s = []

    for meas in dataset.measurements:
        y0 = jnp.array([meas.initial_conditions[species] for species in all_species])
        full_y0s.append(y0)

    y0s = jnp.stack(full_y0s)

    return data, times, y0s, constants


def _configure_simulation_function(
    model: "Model",
    surrogate: Optional[Surrogate],
    dataset: Dataset,
    data: Array,
    y0s: Array,
    times: Array,
    constants: Array,
) -> Tuple[Callable, Array, Array, Array, Array]:
    """Configure the appropriate simulation function based on model type.

    Args:
        model: Model to simulate
        surrogate: Optional surrogate model
        dataset: Dataset for simulation
        data: Observation data array
        y0s: Initial conditions array
        times: Time points array

    Returns:
        Tuple of (simulation function, modified data, modified y0s, modified times)
    """

    model = model._replace_assignments()

    if surrogate is not None:
        # Use surrogate model for rate prediction
        data, times, _ = dataset.to_jax_arrays(model.get_species_order())
        constants = dataset.to_y0_matrix(species_order=model.get_constants_order())
        rate_fun = model._setup_rate_function(in_axes=None)
        y0s = data.reshape(-1, data.shape[-1])

        def sim_func(y0s, theta, constants, times):
            return rate_fun(times, y0s, (theta, constants))

        sim_func = jax.jit(jax.vmap(sim_func, in_axes=(0, None, 0, 0)))
        times = times.ravel()
        constants = jnp.repeat(constants, data.shape[1], axis=0)
        data = surrogate.predict_rates(dataset=dataset)
    else:
        # Create a new simulation function with correct axes for MCMC
        # We need to vmap over measurements (axis 0) for y0s, constants, and times
        # but not over theta (same parameters for all measurements)
        from catalax.tools.simulation import Simulation

        odes = [model.odes[species] for species in model.get_species_order()]
        simulation_setup = Simulation(
            odes=odes,
            parameters=model.get_parameter_order(),
            stoich_mat=model._get_stoich_mat(),
            constants=model.get_constants_order(),
            dt0=0.1,  # Default dt0
            rtol=1e-5,
            atol=1e-5,
            max_steps=64**4,
        )

        # Set up simulation with correct in_axes for MCMC
        sim_func, _ = simulation_setup._prepare_func(in_axes=(0, None, 0, 0))

    return sim_func, data, y0s, times, constants  # type: ignore



import json
import equinox as eqx


def save(filename, hyperparams, model):
    with open(filename, "wb") as f:
        hyperparam_str = json.dumps(hyperparams, default=str)
        f.write((hyperparam_str + "\n").encode())
        eqx.tree_serialise_leaves(f, model)



from typing import Callable, List, Optional, Any, Tuple

import diffrax as dfx
import equinox as eqx
import jax
import jax.numpy as jnp
from diffrax import (
    ConstantStepSize,
    PIDController,
    SaveAt,
    diffeqsolve,
)
from pydantic import BaseModel, PrivateAttr, ConfigDict

from catalax.model.inaxes import InAxes
from catalax.model.ode import ODE
from .symbolicmodule import SymbolicModule

NON_ADAPTIVE_SOLVERS = [
    dfx.Euler,
    dfx.Heun,
]


class Stack(eqx.Module):
    modules: List[eqx.Module]

    def __init__(
        self,
        parameters: List[str],
        odes: List[ODE],
        constants: List[str],
        **kwargs,
    ):
        positionals = {
            "parameters": parameters,
            "constants": constants,
            "states": [str(ode.species.symbol) for ode in odes],
        }

        self.modules = [
            SymbolicModule(ode.equation, positionals=positionals) for ode in odes
        ]

    def __call__(self, t, y, args):
        parameters, constants = args
        rates = jnp.stack(
            [
                module(  # type: ignore
                    t=t,
                    parameters=parameters,
                    states=y,
                    constants=constants,
                )
                for module in self.modules
            ],
            axis=-1,
        )

        return rates


class Simulation(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    odes: List[ODE]
    parameters: List[str]
    constants: List[str]
    stoich_mat: jax.Array
    dt0: float = 0.1
    rtol: float = 1e-5
    atol: float = 1e-5
    max_steps: int = 64**4
    sensitivity: Optional[InAxes] = None
    solver: Any = dfx.Tsit5

    _simulation_func: Optional[Callable] = PrivateAttr(default=None)

    def _prepare_func(self, in_axes=None) -> Tuple[Callable, Stack]:
        """Applies all the necessary transformations to the term and prepares the simulation function"""
        stack = Stack(
            parameters=self.parameters,
            odes=self.odes,
            constants=self.constants,
        )

        controller = self._create_controller()
        simulate_system = self._create_simulate_system(stack, controller)

        if self.sensitivity is not None:
            return self._prepare_sensitivity_func(simulate_system, in_axes)  # type: ignore

        return self._prepare_standard_func(simulate_system, in_axes)  # type: ignore

    def _create_controller(self):
        """Create the appropriate stepsize controller"""
        if self.solver in NON_ADAPTIVE_SOLVERS:
            return ConstantStepSize()
        else:
            return PIDController(rtol=self.rtol, atol=self.atol)

    def _create_simulate_system(self, stack, controller):
        """Create the core simulation function"""

        def _simulate_system(y0, parameters, constants, time):
            sol = diffeqsolve(
                terms=dfx.ODETerm(stack),
                solver=self.solver(),
                t0=0,
                t1=time[-1],
                dt0=self.dt0,
                y0=y0,
                args=(parameters, constants),
                saveat=SaveAt(ts=time),
                stepsize_controller=controller,
                max_steps=self.max_steps,
            )
            return sol.ys

        return _simulate_system

    def _prepare_sensitivity_func(self, simulate_system, in_axes):
        """Prepare simulation function with sensitivity analysis"""
        assert isinstance(self.sensitivity, InAxes), (
            "Expected sensitivity to be an instance of 'InAxes'"
        )

        def sens_fun(y0s, parameters, constants, time):
            return simulate_system(y0s, parameters, constants, time)

        index = self.sensitivity.value.index(0)

        if in_axes is not None:
            return (
                eqx.filter_jit(
                    jax.vmap(jax.jacobian(sens_fun, argnums=index), in_axes=in_axes)
                ),
                self,
            )

        return (
            eqx.filter_jit(jax.jacobian(sens_fun, argnums=int(index))),
            self,
        )

    def _prepare_standard_func(self, simulate_system, in_axes):
        """Prepare standard simulation function"""
        if in_axes is not None:
            return (
                eqx.filter_jit(jax.vmap(simulate_system, in_axes=in_axes)),
                self,
            )
        else:
            return (
                eqx.filter_jit(simulate_system),
                self,
            )

    def __call__(self, y0, parameters, constants, time) -> Any:
        if self._simulation_func is None:
            raise ValueError("Simulation function not initialized")

        return self._simulation_func(y0, parameters, constants, time)



"""Loss functions for model fitting and evaluation.

This module provides loss functions commonly used in optimization and model
evaluation tasks, particularly for scientific computing and experimental
data fitting.
"""

import jax.numpy as jnp
from jax import Array


def l1_loss(y_true: Array, y_pred: Array) -> Array:
    """Calculate the L1 (Manhattan) loss between true and predicted values.

    The L1 loss, also known as Mean Absolute Error (MAE) or Manhattan distance,
    computes the absolute difference between true and predicted values. This loss
    function is robust to outliers compared to L2 loss and is commonly used in
    regression tasks where outlier robustness is desired.

    Args:
        y_true: True/observed values as a JAX array
        y_pred: Predicted values as a JAX array, must have the same shape as y_true

    Returns:
        JAX array containing the element-wise absolute differences between y_true
        and y_pred, with the same shape as the input arrays

    Note:
        This function returns element-wise losses rather than the mean loss.
        To get the mean L1 loss, use jnp.mean(l1_loss(y_true, y_pred)).

    Example:
        >>> import jax.numpy as jnp
        >>> y_true = jnp.array([1.0, 2.0, 3.0])
        >>> y_pred = jnp.array([1.1, 1.8, 3.2])
        >>> loss = l1_loss(y_true, y_pred)
        >>> print(loss)  # [0.1, 0.2, 0.2]
        >>> mean_loss = jnp.mean(loss)
        >>> print(mean_loss)  # 0.16666667
    """
    return jnp.abs(y_true - y_pred)


def mean_absolute_error(y_true: Array, y_pred: Array) -> Array:
    """Calculate the mean absolute error between true and predicted values.

    The mean absolute error computes the average absolute difference between
    true and predicted values. This loss function is robust to outliers compared
    to mean squared error and is commonly used in regression tasks where outlier
    robustness is desired.
    """
    return jnp.mean(jnp.abs(y_true - y_pred))



import re
from typing import Dict

import pandas as pd
from sympy import Derivative, Eq
from dotted_dict import DottedDict

from catalax.model.base import CatalaxBase
from catalax.model.parameter import Parameter


class PrettyDict(DottedDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __repr__(self):
        sorted_items = sorted(self.items(), key=lambda x: x[0])
        df = pd.DataFrame(
            [
                {
                    key: value
                    for key, value in cls.__dict__.items()
                    if key in self._fields_to_print(cls)
                }
                for _, cls in sorted_items
            ]
        )

        try:
            from IPython.display import display

            display(df)

            return ""

        except ImportError:
            return df

    @staticmethod
    def _fields_to_print(cls_: CatalaxBase) -> Dict[str, str]:
        if len(cls_._repr_fields) == 1 and cls_._repr_fields[0] == "__all__":
            return {key: key for key in cls_.__dict__}

        return {key: key for key in cls_._repr_fields}


def odeprint(y, expr):
    """Displays an ODE with a left-hand side derivative"""

    try:
        from IPython.display import display

        display(Eq(Derivative(y, "t"), expr, evaluate=False))

    except ImportError:
        print(f"d{str(y)}_dt' = {str(expr)}")


def eqprint(y, expr):
    """Displays an ODE with a left-hand side derivative"""

    try:
        from IPython.display import display

        display(Eq(y, expr, evaluate=False))

    except ImportError:
        print(f"{str(y)}' = {str(expr)}")


def parameter_exists(name: str, parameters: Dict[str, Parameter]) -> bool:  # type: ignore
    """Checks whether a parameter is already present in a model"""

    return any(param.name == name for param in parameters.values())


def check_symbol(symbol: str) -> None:
    """Checks whether the given symbol is a valid symbol"""

    ERROR_MESSAGE = f"""Symbol '{symbol}' is not a valid symbol. The following rules apply:

    (1) The first character must be a letter
    (2) The remaining characters can be letters and numbers
    (3) The symbol cannot end with an underscore
    (4) The symbol can contain at most one underscore followed by letters and numbers

    These are valid symbols:

    k1, k_12, k_max, k_max1

    """

    # Convert to string to use string methods
    symbol = str(symbol)

    if symbol.endswith("_"):
        raise ValueError(ERROR_MESSAGE)

    pattern = r"""
        ^[A-Za-z]    # First character must be a letter
        [A-Za-z0-9]* # Remaining characters can be letters and numbers
        \_?          # Optional underscore
        [A-Za-z0-9]* # Remaining characters can be letters and numbers
    """

    regex = re.compile(pattern, re.X)

    if not bool(regex.match(symbol)):
        raise ValueError(ERROR_MESSAGE)



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, List, Optional
import ipywidgets as widgets
from IPython.display import display, clear_output
import matplotlib.pyplot as plt
import jax.numpy as jnp

from catalax.dataset.measurement import Measurement

if TYPE_CHECKING:
    from catalax.dataset import Dataset
    from catalax.predictor import Predictor
    from catalax.model.model import Model

# Constants for widget styling and configuration
DEFAULT_PLOT_STYLE = {
    "figsize": (12, 6),
    "linewidth": 2.5,
    "alpha": 0.8,
    "grid": True,
    "grid_alpha": 0.3,
}

DEFAULT_INITIAL_VALUE = 10.0
DEFAULT_TIME_POINTS = 100
MIN_TIME_POINTS = 50
MAX_TIME_POINTS = 1000
CONTROLS_PER_ROW = 4  # Number of controls per row

WIDGET_LAYOUTS = {
    "text_input": widgets.Layout(width="100px"),
    "label": widgets.Layout(width="100px"),
    "control_item": widgets.Layout(
        justify_content="flex-start",
        align_items="center",
        margin="5px 1px",
        width="150px",
    ),
    "control_row": widgets.Layout(
        justify_content="flex-start", align_items="center", margin="5px 0px"
    ),
    "controls_area": widgets.Layout(
        border="2px solid #dee2e6",
        border_radius="10px",
        padding="20px",
        margin="10px 0px",
        background_color="#ffffff",
    ),
    "main_widget": widgets.Layout(
        border="1px solid #ced4da",
        border_radius="15px",
        padding="15px",
        background_color="#f8f9fa",
    ),
}

PLOT_AESTHETICS = {
    "xlabel_fontsize": 12,
    "ylabel_fontsize": 12,
    "legend_fontsize": 10,
    "background_color": "#f8f9fa",
}


class PredictionWidget:
    """
    A class to encapsulate the interactive prediction widget functionality.

    This class separates concerns and makes the widget more maintainable
    by organizing related functionality into methods.
    """

    def __init__(
        self,
        model: Model,
        predictor: Predictor,
        t0: float = 0.0,
        t1: float = 100.0,
        ref_measurement: Optional[Measurement] = None,
    ):
        """
        Initialize the interactive prediction widget.

        Args:
            model: The model containing species information
            predictor: The predictor for generating predictions
            t0: Start time for predictions
            t1: End time for predictions
            ref_measurement: Reference measurement for initial conditions
        """
        self.model = model
        self.predictor = predictor
        self.t0 = t0
        self.t1 = t1
        self.ref_measurement = ref_measurement
        self.plot_style = DEFAULT_PLOT_STYLE

        # Suppress warnings for cleaner output
        warnings.filterwarnings("ignore", category=UserWarning)

        # Initialize widget components
        self._setup_species_info()
        self._setup_default_values()
        self._create_controls()
        self._setup_output()

    def _setup_species_info(self) -> None:
        """Extract species information from the model."""
        self.species_order = self.predictor.get_species_order()
        self.species_names = self._get_species_display_names()

    def _get_species_display_names(self) -> List[str]:
        """
        Get display names for species, preferring name over symbol.

        Returns:
            List of species display names
        """
        names = []
        for species_symbol in self.species_order:
            species_obj = self.model.species[species_symbol]
            name = species_obj.name if species_obj.name is not None else species_symbol
            names.append(name)
        return names

    def _setup_default_values(self) -> None:
        """Setup default values for initial conditions."""
        if self.ref_measurement is not None:
            self.default_values = self.ref_measurement
        else:
            self.default_values = {
                sp: DEFAULT_INITIAL_VALUE for sp in self.species_order
            }

    def _get_initial_value(self, species: str) -> float:
        """
        Get initial value for a species from default values.

        Args:
            species: Species symbol

        Returns:
            Initial concentration value
        """
        if isinstance(self.default_values, Measurement):
            return self.default_values.initial_conditions.get(
                species, DEFAULT_INITIAL_VALUE
            )
        else:
            return self.default_values.get(species, DEFAULT_INITIAL_VALUE)

    def _create_controls(self) -> None:
        """Create all widget controls."""
        self.controls = {}

        # Create species controls
        for species, name in zip(self.species_order, self.species_names):
            initial_val = self._get_initial_value(species)

            text_input = widgets.FloatText(
                value=initial_val,
                description="",
                style={"description_width": "0px"},
                layout=WIDGET_LAYOUTS["text_input"],
            )

            label = widgets.HTML(
                value=f"<b>{name}:</b>", layout=WIDGET_LAYOUTS["label"]
            )

            self.controls[species] = {"text": text_input, "label": label}

        # Create time control
        self.t1_input = widgets.FloatText(
            value=self.t1,
            description="",
            style={"description_width": "0px"},
            layout=WIDGET_LAYOUTS["text_input"],
        )

        self.t1_label = widgets.HTML(
            value="<b>End Time (t1):</b>", layout=WIDGET_LAYOUTS["label"]
        )

    def _setup_output(self) -> None:
        """Setup the output widget for plots."""
        self.output = widgets.Output()

    def _create_prediction_dataset(self) -> Dataset:
        """
        Create a dataset with current input values for prediction.

        Returns:
            Dataset configured for prediction

        Raises:
            ValueError: If input values are invalid
        """

        from catalax.dataset.dataset import Dataset

        try:
            prediction_dataset = Dataset.from_model(self.model)

            # Build initial conditions from text input values
            initial_conditions = {
                species: self.controls[species]["text"].value
                for species in self.species_order
            }

            # Validate initial conditions
            for species, value in initial_conditions.items():
                if value < 0:
                    raise ValueError(
                        f"Initial condition for {species} cannot be negative: {value}"
                    )

            prediction_dataset.add_initial(**initial_conditions)

            # Setup time points
            current_t1 = max(self.t1_input.value, self.t0 + 0.1)  # Ensure t1 > t0
            time_points = min(
                max(DEFAULT_TIME_POINTS, MIN_TIME_POINTS), MAX_TIME_POINTS
            )
            prediction_dataset.measurements[0].time = jnp.linspace(
                self.t0, current_t1, time_points
            )

            return prediction_dataset

        except Exception as e:
            raise ValueError(f"Failed to create prediction dataset: {str(e)}")

    def _generate_predictions(self, dataset: Dataset) -> Dataset:
        """
        Generate predictions using the predictor.

        Args:
            dataset: Input dataset for prediction

        Returns:
            Dataset with predictions

        Raises:
            RuntimeError: If prediction fails
        """
        try:
            return self.predictor.predict(dataset)
        except Exception as e:
            raise RuntimeError(f"Prediction failed: {str(e)}")

    def _create_plot(self, predictions: Dataset) -> None:
        """
        Create and display the prediction plot.

        Args:
            predictions: Dataset containing prediction results
        """
        fig, ax = plt.subplots(figsize=self.plot_style["figsize"])

        # Plot predictions with species names as labels
        for i, species_symbol in enumerate(self.species_order):
            species_name = self.species_names[i]
            time_data = predictions.measurements[0].time
            concentration_data = predictions.measurements[0].data[species_symbol]

            ax.plot(
                time_data,  # type: ignore
                concentration_data,  # type: ignore
                label=species_name,
                linewidth=self.plot_style["linewidth"],
                alpha=self.plot_style["alpha"],
            )

        # Enhance plot aesthetics
        ax.set_xlabel(
            "Time", fontsize=PLOT_AESTHETICS["xlabel_fontsize"], fontweight="bold"
        )
        ax.set_ylabel(
            "Concentration",
            fontsize=PLOT_AESTHETICS["ylabel_fontsize"],
            fontweight="bold",
        )

        if self.plot_style["grid"]:
            ax.grid(True, alpha=self.plot_style["grid_alpha"], linestyle="--")

        # Place legend outside the plot area
        ax.legend(
            bbox_to_anchor=(1.05, 1),
            loc="upper left",
            frameon=True,
            fancybox=True,
            shadow=True,
            fontsize=PLOT_AESTHETICS["legend_fontsize"],
        )

        # Set background color
        ax.set_facecolor(PLOT_AESTHETICS["background_color"])

        plt.tight_layout()
        plt.show()

    def _update_plot(self, change=None) -> None:
        """
        Update the prediction plot based on current input values.

        Args:
            change: Widget change event (unused)
        """
        with self.output:
            clear_output(wait=True)

            try:
                # Create dataset and generate predictions
                prediction_dataset = self._create_prediction_dataset()
                predictions = self._generate_predictions(prediction_dataset)

                # Create and display plot
                self._create_plot(predictions)

            except (ValueError, RuntimeError) as e:
                # Display error message instead of crashing
                print(f"Error: {str(e)}")
                print("Please check your input values and try again.")

    def _create_controls_area(self) -> widgets.VBox:
        """
        Create the controls area containing all controls arranged in rows.

        Returns:
            Controls area widget containing all controls in horizontal rows
        """
        # Create control items for species
        control_items = []
        for species in self.species_order:
            item = widgets.VBox(
                [self.controls[species]["label"], self.controls[species]["text"]],
                layout=WIDGET_LAYOUTS["control_item"],
            )
            control_items.append(item)

        # Add time control item
        time_item = widgets.VBox(
            [self.t1_label, self.t1_input],
            layout=WIDGET_LAYOUTS["control_item"],
        )
        control_items.append(time_item)

        # Create rows with specified number of controls per row
        control_rows = []
        for i in range(0, len(control_items), CONTROLS_PER_ROW):
            row_items = control_items[i : i + CONTROLS_PER_ROW]
            row = widgets.HBox(
                row_items,
                layout=WIDGET_LAYOUTS["control_row"],
            )
            control_rows.append(row)

        # Create controls area
        controls_area = widgets.VBox(
            control_rows,
            layout=WIDGET_LAYOUTS["controls_area"],
        )

        return controls_area

    def _setup_observers(self) -> None:
        """Setup observers for all input widgets."""
        # Set up species input observers
        for species in self.species_order:
            self.controls[species]["text"].observe(self._update_plot, names="value")

        # Set up time input observer
        self.t1_input.observe(self._update_plot, names="value")

    def create_widget(self) -> widgets.Widget:
        """
        Create and return the complete interactive widget.

        Returns:
            Complete interactive widget ready for display
        """
        # Create controls area
        controls_area = self._create_controls_area()

        # Setup observers
        self._setup_observers()

        # Create main layout
        main_content = widgets.VBox(
            [controls_area, self.output],
            layout=widgets.Layout(justify_content="flex-start", align_items="stretch"),
        )

        # Main widget container
        main_widget = widgets.VBox(
            [main_content],
            layout=WIDGET_LAYOUTS["main_widget"],
        )

        # Display the widget and generate initial plot
        display(main_widget)
        self._update_plot()

        return main_widget


def create_interactive_prediction_widget(
    model: Model,
    predictor: Predictor,
    t0: float = 0.0,
    t1: float = 100.0,
    ref_measurement: Optional[Measurement] = None,
) -> widgets.Widget:
    """
    Create an interactive widget for exploring Neural ODE predictions with adjustable initial conditions.

    This function creates a comprehensive interactive interface that allows users to:
    - Adjust initial conditions for each species in the model
    - Modify the end time for predictions
    - View real-time updates of the prediction plot

    Args:
        model: The model containing species information and structure
        predictor: The predictor object for generating predictions (e.g., trained Neural ODE)
        t0: Start time for predictions (default: 0.0)
        t1: End time for predictions (default: 100.0)
        ref_measurement: Reference measurement containing initial conditions to use as defaults.
                        If None, uses default values of 10.0 for all species
        plot_style: Dictionary of matplotlib styling options. Merged with defaults:
                   - figsize: (12, 6)
                   - linewidth: 2.5
                   - alpha: 0.8
                   - grid: True
                   - grid_alpha: 0.3

    Returns:
        Interactive widget with text inputs and plot output. The widget automatically
        displays upon creation and updates plots in real-time as inputs change.

    Raises:
        ValueError: If the model or predictor are invalid
        RuntimeError: If widget creation fails

    Note:
        This function is designed for use in Jupyter notebooks and requires
        ipywidgets and matplotlib for proper functionality.
    """
    try:
        widget_creator = PredictionWidget(
            model=model,
            predictor=predictor,
            t0=t0,
            t1=t1,
            ref_measurement=ref_measurement,
        )
        return widget_creator.create_widget()

    except Exception as e:
        raise RuntimeError(f"Failed to create interactive prediction widget: {str(e)}")



from typing import List

from pydantic import PrivateAttr
from sympy import Expr

from catalax.model.base import CatalaxBase


class Constant(CatalaxBase):
    name: str
    symbol: Expr

    _repr_fields: List[str] = PrivateAttr(
        default=[
            "name",
            "symbol",
        ]
    )



import json
from typing import List, Optional

import jax
import diffrax
import jax.numpy as jnp
import equinox as eqx
import jax.random as jrandom
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors

from catalax import Model
from catalax.dataset.dataset import Dataset
from catalax.tools import Stack
from .neuralbase import NeuralBase


class UniversalODE(NeuralBase):
    parameters: jax.Array
    vector_field: Stack
    alpha_residual: jax.Array
    gate_matrix: jax.Array
    use_gate: bool = eqx.field(static=True)
    model: dict = eqx.field(static=True)

    def __init__(
        self,
        data_size: int,
        width_size: int,
        depth: int,
        model: Model | str | dict,
        observable_indices: List[int],
        solver=diffrax.Tsit5,
        use_final_bias: bool = False,
        activation=jax.nn.softplus,
        final_activation=None,
        use_gate: bool = True,
        *,
        key,
        **kwargs,
    ):
        if isinstance(model, str):
            model = Model.from_dict(json.loads(model))
        elif isinstance(model, dict):
            model = Model.from_dict(model)

        mlp_key, alpha_key, gate_key = jrandom.split(key, 3)
        species_order = model.get_species_order()
        super().__init__(
            data_size=data_size,
            width_size=width_size,
            depth=depth,
            species_order=species_order,
            observable_indices=observable_indices,
            solver=solver,
            activation=activation,
            key=mlp_key,
            use_final_bias=use_final_bias,
            final_activation=final_activation,
            model=model.to_dict(),
        )

        if use_gate and len(species_order) > 1:
            self.use_gate = use_gate
        else:
            self.use_gate = True

        self.model = model.to_dict()
        self.parameters = jnp.array(model._get_parameters())
        self.vector_field = eqx.filter_jit(model._setup_rate_function())  # type: ignore
        self.alpha_residual = 0.05 * jrandom.normal(alpha_key, (len(species_order),))
        self.gate_matrix = 0.01 * jrandom.normal(
            gate_key,
            (
                len(species_order),
                len(species_order),
            ),
        )

    def _corrective_term(self, t, y, args):
        """Get the corrective term of the model at the given states.

        The corrective term is the neural network contribution to the rates.
        """
        gate = jax.nn.sigmoid(10 * (self.gate_matrix @ y))
        alpha_residual = jax.nn.softplus(self.alpha_residual)
        ann_rates = self.func(t, y, args)
        return alpha_residual * gate * ann_rates

    def _combined_term(self, t, y, args):
        """Merges neural network terms and vector field"""
        mech_rates = self.vector_field(t, y, (self.parameters, None))

        if not self.use_gate:
            return mech_rates + self.func(t, y, args)

        return mech_rates + self._corrective_term(t, y, args)

    def __call__(self, ts, y0):
        solution = diffrax.diffeqsolve(
            diffrax.ODETerm(self._combined_term),  # type: ignore
            self.solver(),  # type: ignore
            t0=0.0,  # type: ignore
            t1=ts[-1],
            dt0=ts[1] - ts[0],
            y0=y0,
            saveat=diffrax.SaveAt(ts=ts),  # type: ignore
            max_steps=64**4,
        )
        return solution.ys

    def corrective_term(self, dataset: Dataset) -> tuple[jax.Array, jax.Array]:
        """Get the corrective term of the model at the given states.

        The corrective term is the neural network contribution to the rates.
        """

        data, times, _ = dataset.to_jax_arrays(species_order=self.species_order)
        y = data.reshape(-1, data.shape[-1])
        t = times.ravel()

        corrections = jax.vmap(self._corrective_term, in_axes=(0, 0, None))(t, y, None)
        return corrections.reshape(data.shape), times

    def plot_corrections_over_time(
        self,
        dataset: Dataset,
        show: bool = True,
        path: Optional[str] = None,
        n_steps: int = 100,
        figsize: tuple[float, float] = (10, 3.5),
    ) -> Optional[Figure]:
        # Calculate proper number of rows for odd/even measurements
        uode_pred = self.predict(dataset, n_steps=n_steps)
        corrections, times = self.corrective_term(uode_pred)
        n_meas, _, n_species = corrections.shape

        width, height = figsize

        if n_species == 1:
            # If there is only one species, we can plot in a grid of (n_measurements // 2, 2)
            fig, ax = plt.subplots(
                (n_meas + 1) // 2,
                2,
                figsize=(width, height * ((n_meas + 1) // 2)),
            )
        else:
            fig, ax = plt.subplots(
                n_meas,
                n_species,
                figsize=(width * 0.5 * n_species, height * 0.7 * n_meas),
            )

        ax = ax.flatten()

        color_iter = iter(mcolors.TABLEAU_COLORS)
        species_names = [(species, next(color_iter)) for species in dataset.species]

        ax_index = 0
        for i in range(n_meas):
            corrective_term = corrections[i]

            for j in range(n_species):
                name, color = species_names[j]
                ax[ax_index].plot(
                    times[i],
                    corrective_term[:, j],
                    color=color,
                    linewidth=2,
                    label="Correction",
                )

                ax[ax_index].set_title(f"{name} (Meas. {i + 1})", fontsize=12)
                ax[ax_index].set_xlabel("Time", fontsize=10)
                ax[ax_index].set_ylabel("Corrective Term", fontsize=10)
                ax[ax_index].grid(True, which="both")
                ax[ax_index].grid(True, which="minor", alpha=0.3, linestyle="--")
                ax[ax_index].minorticks_on()
                ax[ax_index].legend(fontsize=10)
                ax[ax_index].set_ylim(
                    jnp.min(corrections[:, :, j]) * 1.2,
                    max(float(jnp.max(corrections[:, :, j]) * 1.2), 0.5),
                )

                ax_index += 1

        # Hide unused subplots when odd number of measurements
        total_subplots = len(ax)
        for i in range(n_meas * n_species, total_subplots):
            ax[i].set_visible(False)

        plt.tight_layout()

        if path:
            plt.savefig(path)

        if show:
            return plt.show()

        return fig

    def plot_corrections_over_input(
        self,
        dataset: Dataset,
        show: bool = True,
        path: Optional[str] = None,
        figsize: tuple[float, float] = (10, 3.5),
    ) -> Optional[Figure]:
        uode_pred = self.predict(dataset, use_times=True)
        data, _, _ = dataset.to_jax_arrays(species_order=self.species_order)
        corrections, _ = self.corrective_term(uode_pred)

        # Unravel data and corrections
        data = data.reshape(-1, data.shape[-1])
        corrections = corrections.reshape(-1, corrections.shape[-1])

        # Get number of species and colors
        n_species = data.shape[-1]
        width, height = figsize
        fig, ax = plt.subplots(n_species, 1, figsize=(width, height * n_species))

        if n_species == 1:
            ax = [ax]

        color_iter = iter(mcolors.TABLEAU_COLORS)
        species_names = [(species, next(color_iter)) for species in dataset.species]

        for i, (species, color) in enumerate(species_names):
            ax[i].scatter(
                data[:, i],
                corrections[:, i],
                color=color,
            )
            ax[i].set_xlabel("Concentration", fontsize=10)
            ax[i].set_ylabel("Corrective Term", fontsize=10)
            ax[i].set_title(f"{species}", fontsize=12)
            ax[i].grid(True, which="both")
            ax[i].grid(True, which="minor", alpha=0.3, linestyle="--")
            ax[i].minorticks_on()

        # Hide unused subplots when odd number of species
        if n_species > 1:
            total_subplots = len(ax)
            for i in range(n_species, total_subplots):
                ax[i].set_visible(False)

        plt.tight_layout()

        if path:
            plt.savefig(path)

        if show:
            return plt.show()

        return fig

    def gate_activation(self, y: jax.Array) -> jax.Array:
        """Get the gate of the model at the given states.

        The gate is a smooth function that determines if the neural network
        contribution is active or not. The gate is triggered by state concentrations
        that flow into a sigmoid function. The row weights of the gate matrix determine
        the sensitivity of the gate to each state. weights can cancel out each other and
        thus cross-talk between states can be controlled.

        Args:
            y: jax.Array of shape (n_species,) or (n_time, n_species)

        Returns:
            Gate
        """

        def _get_gate(y: jax.Array) -> jax.Array:
            return jax.nn.sigmoid(10 * (self.gate_matrix @ y))

        if y.ndim == 1:
            return _get_gate(y)
        elif y.ndim == 2:
            return jax.vmap(_get_gate)(y)
        else:
            raise ValueError("y must be a 1D or 2D array")

    def get_alpha_residual(self) -> jax.Array:
        """Get the alpha residual of the model.

        The alpha residual is a scalar that scales the neural network contribution.
        This way and in combination with L1 regularization can introduce sparsity
        in the neural network contribution. This is particularly useful for symbolic
        regression to recover corrective terms.

        Returns:
            Alpha residual
        """
        return jax.nn.softplus(self.alpha_residual)

    def get_rates(
        self,
        t: jax.Array,
        y: jax.Array,
    ) -> jax.Array:
        """Get the rates of the model.

        This basically evaluates the MLP at the given time points and states, which
        is useful for quiver plots and MCMC surrogates.

        Args:
            t: Time points
            y: States

        Returns:
            Rates
        """
        t, y, _ = self._validate_rate_input(t, y, None)
        res: jax.Array = jax.vmap(self._combined_term, in_axes=(0, 0, None))(t, y, None)
        return res



from .dataset import Dataset
from .measurement import Measurement

import arviz as az  # noqa: F401

# Set plotting style
az.style.use("arviz-doc")  # type: ignore

__all__ = [
    "Dataset",
    "Measurement",
]



from typing import List
from pydantic import BaseModel, PrivateAttr, ConfigDict


class CatalaxBase(BaseModel):
    model_config = ConfigDict(
        use_enum_values=True,
        arbitrary_types_allowed=True,
        validate_assignment=True,
    )

    _repr_fields: List[str] = PrivateAttr(default=["__all__"])
